// 
//  Copyright 2008 OCP-IP
//
// ============================================================================
//      Project : OCP SLD WG
//       Author : Anssi Haverinen, Nokia Inc.
//                Robert Guenzel (from TU of Braunschweig) for Greensocs Ltd.
//           $Id: master_r.cc,v 1.1 2005/01/07 03:42:33 Anssi Exp $
//
//  Description : OCP API - TL1 profile example
// ============================================================================

#include "master_r.h"

// ----------------------------------------------------------------------------
// Constructor
// ----------------------------------------------------------------------------
Master_r::Master_r (sc_core::sc_module_name name_): 
  sc_core::sc_module (name_), ipP("ipPort", ocpip::ocp_master_socket_tl1<32>::mm_txn_with_data()), ongoing_req(false), resp(NULL) {

  // initialize common members
  cnt = 0;
  wraplen = 0;
  addr = 0;
  burstcnt = 0;
  SC_METHOD(proc);
  sensitive<<clk.pos(); 
  dont_initialize();
  
  ipP.register_nb_transport_bw(this, &Master_r::nb_transport);
  ipP.activate_delta_cycle_protection();
  rspcnt=0;
}

// ----------------------------------------------------------------------------
// Destructor
// ----------------------------------------------------------------------------
Master_r::~Master_r() {}

// ----------------------------------------------------------------------------
//  Method : Master_r::proc()
//
//  Synchronous Master process
//
// ----------------------------------------------------------------------------
void Master_r::proc() {

  // check if request channel is free
  if (!ongoing_req) {
    if (cnt==0 | cnt==8 | cnt==16 | (cnt>23 && cnt<28)){
      req=ipP.get_transaction();
      ipP.reserve_data_size(*req, 8*sizeof(Td));
      req->set_command(tlm::TLM_READ_COMMAND);
      ocpip::burst_sequence* b_seq;
      ipP.get_extension<ocpip::burst_sequence>(b_seq, *req);
      req->set_streaming_width(8*sizeof(Td));
      req->set_address(addr);
      if (cnt==0)
        b_seq->value.sequence=ocpip::INCR;
      else
      if (cnt==8){
        b_seq->value.sequence=ocpip::WRAP;
        b_seq->value.xor_wrap_address=addr; //wrap=incr in this example
        req->set_address(addr); //first burst went from 0 to 31, so that one starts at 32
      }
      else
      if (cnt==16){
        b_seq->value.sequence=ocpip::STRM;
        req->set_streaming_width(sizeof(Td));
      }
      else
        b_seq->value.sequence=ocpip::INCR;
      ipP.validate_extension<ocpip::burst_sequence>(*req);
      ocpip::burst_length* b_len;
      ipP.get_extension<ocpip::burst_length>(b_len, *req);
      b_len->value=8;
      ipP.validate_extension<ocpip::burst_length>(*req);
      req->set_byte_enable_ptr(NULL);
      req->set_response_status(tlm::TLM_INCOMPLETE_RESPONSE);
    }
    
    if (cnt<8) {
      wraplen = 0;
      phase=tlm::BEGIN_REQ;
      // Set OCP command type
      if (burstcnt==7) {
        burstcnt=0;
      }
      else {
        burstcnt++;
      }	
    }
    else if (cnt<16) {
      wraplen = 8;
      phase=tlm::BEGIN_REQ;
      req->set_address(addr);
      if (burstcnt==7) {
        burstcnt=0;
      }
      else {
        burstcnt++;
      }
    }
    else if (cnt<24){
      wraplen = 0;
      phase=tlm::BEGIN_REQ;
      req->set_address(4);
      if (burstcnt==7) {
        burstcnt=0;
      }
      else {
        burstcnt++;
      }
    }
    else if (cnt<28){ // SRMD burst
      wraplen = 0;
      phase=tlm::BEGIN_REQ;
      ipP.validate_extension<ocpip::srmd>(*req);
      req->set_address(4);
    }
    else {
      phase=tlm::UNINITIALIZED_PHASE;
    }

    // Send request
    if (phase != tlm::UNINITIALIZED_PHASE){
      time=sc_core::SC_ZERO_TIME;
      tlm::tlm_sync_enum retVal = ipP->nb_transport_fw(*req, phase, time);
      switch(retVal){
        case tlm::TLM_ACCEPTED: ongoing_req=true; break;
        case tlm::TLM_UPDATED: 
          switch (phase){
            case tlm::END_REQ: 
              /*
              if (cnt>23){
                std::cerr<<"Error in "<<name()<<" : BEGIN_SRMD_REQ was updated with "<<phase<<std::endl;
                exit(1);
              }
              */
              break;
            default:
            /*
              if (phase==ocpip::END_SRMD_REQ){
                if (cnt<28){
                  std::cerr<<"Error in "<<name()<<" : BEGIN_REQ was updated with "<<phase<<std::endl;
                  exit(1);
                }
              }
              else{
              */
                std::cerr<<"Error in "<<name()<<" : got unexpected phase update to "<<phase<<" in response to BEGIN_(SRMD_)REQ."<<std::endl;
                exit(1);                
              //}
          }
          break;
        case tlm::TLM_COMPLETED:;
      }


#ifdef DEBUG_G1
      std::cout << "Master_r sent request "
	   << "  time  = " << sc_core::sc_time_stamp().to_seconds()
	   << "  MAddr  = " << req->get_address() << std::endl;
#endif
      if (req->get_command() == tlm::TLM_READ_COMMAND) {
        if (cnt>23) {//srmd
          for (int i=0; i<8; i++)
            expected_rsps.push_back(expected_rsp(req, i, i==7));
        }
        else
          expected_rsps.push_back(expected_rsp(req, (((burstcnt)?burstcnt:8)-1), burstcnt==0));
      }

      // Address generator
      if (wraplen) {
        if ((addr+4) < ((addr & (~(wraplen-1)*4))+ wraplen*4))
          addr = addr+4; 
        else
          addr = addr & (~((wraplen-1)*4));
      }
      else
        if (addr < 124)
          addr += 4;
        else
          addr = 0;    
      cnt++;
    }
  }

  if (resp) {
    time=sc_core::SC_ZERO_TIME;
    phase=tlm::END_RESP;
    ipP->nb_transport_fw(*resp, phase, time);
    assert(expected_rsps.front().txn==resp);
    if (resp->get_response_status() == tlm::TLM_OK_RESPONSE) {
#ifdef DEBUG_G1
      std::cout << "Master got valid response "
       << "  time  = " << sc_core::sc_time_stamp().to_seconds()
       << "  SResp  = " << resp->get_response_status()
       << "  SData  = " << (*(((Td*)(resp->get_data_ptr()))+expected_rsps.front().chunk)) << std::endl;
#endif
      if (expected_rsps.front().last) {
        ipP.release_transaction(resp); //we are done with this txn now
      }
      resp=NULL;
      expected_rsps.pop_front();
    }
  }
} // end of method


tlm::tlm_sync_enum Master_r::nb_transport(tlm::tlm_generic_payload& txn, tlm::tlm_phase& ph, sc_core::sc_time& tim) {
  switch(ph){
    case tlm::END_REQ:
      /*
      if (cnt>24){
        std::cerr<<"Error in "<<name()<<" : BEGIN_SRMD_REQ was finished with "<<phase<<std::endl;
        exit(1);
      }
      */
      ongoing_req=false;
      break;
    case tlm::BEGIN_RESP: resp=&txn; break;
    default:
    /*
      if (ph==ocpip::END_SRMD_REQ){
        if (cnt<25){
          std::cerr<<"Error in "<<name()<<" : BEGIN_REQ was finished with "<<phase<<std::endl;
          exit(1);
        }      
        ongoing_req=false;
      }
      else{
      */
        std::cerr<<"Error in "<<name()<<" : got unexpected phase "<<ph<<" in nb_transport_bw"<<std::endl;
        exit(1);
      //}
  }
  return tlm::TLM_ACCEPTED;
}

