// 
//  (c) Copyright OCP-IP 2003, 2004
//  OCP-IP Confidential and Proprietary
//
// ============================================================================
//      Project : OCP SLD WG
//       Author : Norman Weyrich, Synopsys Inc.
//                Alan Kamas, for Sonics Inc.
//                Yann Bajot, Prosilog
//                Anssi Haverinen, Nokia Inc.
//         $Id: tl_master_if.h,v 1.3 2004/09/05 17:57:56 halexan Exp $
//
//  Description : Transaction Level Master interface declaration.
//                These methods are implemented in the Channel
//                (see tl_channel.h)
//
//  Change History
//             12/14/2003 Original revision for OCP 2.0
//             02/03/2004 Added MunreleasePE()
// ============================================================================//

#ifndef _TL_MASTER_IF
#define _TL_MASTER_IF

#include "systemc.h"

// forward declaration for parameter class
template <class TdataCl> class ParamCl;

#include "tl_comm_cl.h"
#include "tl_direct_if.h"

template <class TdataCl> class TLmasterIF
  : public MdirectIF<TdataCl>
{
  public:

  typedef typename TdataCl::DataType Td;
  typedef typename TdataCl::AddrType Ta;

  // Methods to initiate requests and get responses.
  virtual void Mrelease() = 0;
  virtual void MreleasePE() = 0;
  virtual void MunreleasePE() = 0;
  virtual void Mrelease(sc_time) = 0;
  virtual bool MputRequest() = 0;
  virtual bool MputWriteRequest() = 0;
  virtual bool MputWriteRequestBlocking() = 0;
  virtual bool MputReadRequest() = 0;
  virtual bool MputReadRequestBlocking() = 0;
  virtual bool MgetResponse(bool) = 0;
  virtual bool MgetResponseBlocking(bool) = 0;
  virtual bool MgetResponsePE() = 0;
  virtual bool MgetSbusy() const = 0;

  // Second request channel (data handshake)
  virtual bool MputDataRequest() = 0;
  virtual bool MputDataRequestBlocking() = 0;
  virtual bool MgetSbusyData() const = 0;

  // This event is triggered if an Mput or an Sput method is called
  // in the Channel connected to the Master.
  virtual const sc_event& default_event() const = 0;

  // Access methods to the user defined data structures of the Channel.
  virtual TdataCl          *GetDataCl()  = 0;
  virtual ParamCl<TdataCl> *GetParamCl() = 0;
  virtual CommCl           *GetCommCl()  = 0;

  // Direct access method Master to Slave direction.
  virtual bool MputDirect(int, bool, Td*, Ta, int) = 0;
  virtual void MregisterDirectIF(MdirectIF<TdataCl>*) = 0;

  // Reset methods
  virtual void reset() = 0;
  virtual void remove_reset() = 0;
  virtual bool get_reset() = 0;

};

#endif // _TL_MASTER_IF
